@@ -14,6 +14,7 @@ use std::process::Command;
 
 use anyhow::{bail, Context, Result};
 use tempfile::TempDir;
+use toml::{Table, Value};
 use walkdir::WalkDir;
 
 /// The name of the profile used for buliding the sysroot.
@@ -216,20 +217,12 @@ impl<'a> SysrootBuilder<'a> {
     }
 
     /// Appends the given flag.
-    ///
-    /// If no `--cap-lints` argument is configured, we will add `--cap-lints=warn`.
-    /// This emulates the usual behavior of Cargo: Lints are normally capped when building
-    /// dependencies, except that they are not capped when building path dependencies, except that
-    /// path dependencies are still capped if they are part of `-Zbuild-std`.
     pub fn rustflag(mut self, rustflag: impl Into<OsString>) -> Self {
         self.rustflags.push(rustflag.into());
         self
     }
 
     /// Appends the given flags.
-    ///
-    /// If no `--cap-lints` argument is configured, we will add `--cap-lints=warn`. See
-    /// [`SysrootBuilder::rustflag`] for more explanation.
     pub fn rustflags(mut self, rustflags: impl IntoIterator<Item = impl Into<OsString>>) -> Self {
         self.rustflags.extend(rustflags.into_iter().map(Into::into));
         self
@@ -306,46 +299,50 @@ impl<'a> SysrootBuilder<'a> {
 
     /// Generate the contents of the manifest file for the sysroot build.
     fn gen_manifest(&self, src_dir: &Path) -> String {
-        let have_sysroot_crate = src_dir.join("sysroot").exists();
+        let have_builtins_crate = src_dir.join("compiler-builtins").exists();
         let crates = match &self.config {
+            // As long as the sysroot has a patch for compiler_builtins, we need to include it in the
+            // NoStd variant explicitly.
+            SysrootConfig::NoStd if have_builtins_crate => format!(
+                r#"
+                [dependencies.core]
+                path = {src_dir_core:?}
+                [dependencies.alloc]
+                path = {src_dir_alloc:?}
+                [dependencies.compiler_builtins]
+                path = {src_dir_builtins:?}
+                features = ["rustc-dep-of-std", "mem"]
+                "#,
+                src_dir_core = src_dir.join("core"),
+                src_dir_alloc = src_dir.join("alloc"),
+                src_dir_builtins = src_dir.join("compiler-builtins").join("compiler-builtins"),
+            ),
+            // Fallback for old rust without in-tree `compiler-builtins`.
             SysrootConfig::NoStd => format!(
                 r#"
-[dependencies.core]
-path = {src_dir_core:?}
-[dependencies.alloc]
-path = {src_dir_alloc:?}
-[dependencies.compiler_builtins]
-features = ["rustc-dep-of-std", "mem"]
-version = "*"
+                [dependencies.core]
+                path = {src_dir_core:?}
+                [dependencies.alloc]
+                path = {src_dir_alloc:?}
+                [dependencies.compiler_builtins]
+                version = "*"
+                features = ["rustc-dep-of-std", "mem"]
                 "#,
                 src_dir_core = src_dir.join("core"),
                 src_dir_alloc = src_dir.join("alloc"),
             ),
-            SysrootConfig::WithStd { std_features } if have_sysroot_crate => format!(
+            SysrootConfig::WithStd { std_features } => format!(
                 r#"
-[dependencies.std]
-features = {std_features:?}
-path = {src_dir_std:?}
-[dependencies.sysroot]
-path = {src_dir_sysroot:?}
+                [dependencies.std]
+                features = {std_features:?}
+                path = {src_dir_std:?}
+                [dependencies.sysroot]
+                path = {src_dir_sysroot:?}
                 "#,
                 std_features = std_features,
                 src_dir_std = src_dir.join("std"),
                 src_dir_sysroot = src_dir.join("sysroot"),
             ),
-            // Fallback for old rustc where the main crate was `test`, not `sysroot`
-            SysrootConfig::WithStd { std_features } => format!(
-                r#"
-[dependencies.std]
-features = {std_features:?}
-path = {src_dir_std:?}
-[dependencies.test]
-path = {src_dir_test:?}
-                "#,
-                std_features = std_features,
-                src_dir_std = src_dir.join("std"),
-                src_dir_test = src_dir.join("test"),
-            ),
         };
 
         // If we include a patch for rustc-std-workspace-std for no_std sysroot builds, we get a
@@ -354,52 +351,59 @@ path = {src_dir_test:?}
         // Currently the only user of rustc-std-workspace-alloc is std_detect, which is only used
         // by std. So we only need to patch rustc-std-workspace-core in no_std sysroot builds, or
         // that patch also produces a warning.
-        let patches = match &self.config {
-            SysrootConfig::NoStd => format!(
-                r#"
-[patch.crates-io.rustc-std-workspace-core]
-path = {src_dir_workspace_core:?}
-                "#,
-                src_dir_workspace_core = src_dir.join("rustc-std-workspace-core"),
-            ),
-            SysrootConfig::WithStd { .. } => format!(
-                r#"
-[patch.crates-io.rustc-std-workspace-core]
-path = {src_dir_workspace_core:?}
-[patch.crates-io.rustc-std-workspace-alloc]
-path = {src_dir_workspace_alloc:?}
-[patch.crates-io.rustc-std-workspace-std]
-path = {src_dir_workspace_std:?}
-                "#,
-                src_dir_workspace_core = src_dir.join("rustc-std-workspace-core"),
-                src_dir_workspace_alloc = src_dir.join("rustc-std-workspace-alloc"),
-                src_dir_workspace_std = src_dir.join("rustc-std-workspace-std"),
-            ),
+        let unneeded_patches = match &self.config {
+            SysrootConfig::NoStd => &["rustc-std-workspace-alloc", "rustc-std-workspace-std"][..],
+            SysrootConfig::WithStd { .. } => &[][..],
         };
 
-        format!(
+        let mut patches = extract_patches(src_dir);
+        for (repo, repo_patches) in &mut patches {
+            let repo_patches = repo_patches
+                .as_table_mut()
+                .unwrap_or_else(|| panic!("source `{}` is not a table", repo));
+
+            // Remove any patches that we don't need
+            for krate in unneeded_patches {
+                repo_patches.remove(*krate);
+            }
+
+            // Remap paths to be relative to the source directory.
+            for (krate, patch) in repo_patches {
+                if let Some(path) = patch.get_mut("path") {
+                    let curr_path = path
+                        .as_str()
+                        .unwrap_or_else(|| panic!("`{}.path` is not a string", krate));
+
+                    *path = Value::String(src_dir.join(curr_path).display().to_string());
+                }
+            }
+        }
+
+        let mut table: Table = toml::from_str(&format!(
             r#"
-[package]
-authors = ["rustc-build-sysroot"]
-name = "custom-local-sysroot"
-version = "0.0.0"
-edition = "2018"
+            [package]
+            authors = ["rustc-build-sysroot"]
+            name = "custom-local-sysroot"
+            version = "0.0.0"
+            edition = "2018"
 
-[lib]
-# empty dummy, just so that things are being built
-path = "lib.rs"
+            [lib]
+            # empty dummy, just so that things are being built
+            path = "lib.rs"
 
-[profile.{DEFAULT_SYSROOT_PROFILE}]
-# We inherit from the local release profile, but then overwrite some
-# settings to ensure we still get a working sysroot.
-inherits = "release"
-panic = 'unwind'
+            [profile.{DEFAULT_SYSROOT_PROFILE}]
+            # We inherit from the local release profile, but then overwrite some
+            # settings to ensure we still get a working sysroot.
+            inherits = "release"
+            panic = 'unwind'
 
-{crates}
-
-{patches}
+            {crates}
             "#
-        )
+        ))
+        .expect("failed to parse toml");
+
+        table.insert("patch".to_owned(), patches.into());
+        toml::to_string(&table).expect("failed to serialize to toml")
     }
 
     /// Build the `self` sysroot from the given sources.
@@ -435,6 +439,19 @@ panic = 'unwind'
             when_build_required();
         }
 
+        // Create the *parent* directroy of what we are going to create, so that we can later move
+        // into it.
+        fs::create_dir_all(&sysroot_target_dir.parent().unwrap())
+            .context("failed to create target directory")?;
+        // Remove potentially outdated files. Do this via rename to make it atomic.
+        // We do this *before* the step that takes all the time. That means if a bunch of
+        // these builds happen concurrently, then almost certainly this cleanup will happen before
+        // any of them is done, i.e., we only delete outdated files, not new files.
+        // (The delete will happen automatically when `unstaging_dir` gets dropped.)
+        let unstaging_dir =
+            TempDir::new_in(&self.sysroot_dir).context("failed to create un-staging dir")?;
+        let _ = fs::rename(&sysroot_target_dir, &unstaging_dir); // rename may fail if the dir does not exist yet
+
         // Prepare a workspace for cargo
         let build_dir = TempDir::new().context("failed to create tempdir")?;
         // Cargo.lock
@@ -480,8 +497,10 @@ panic = 'unwind'
         // Set rustflags.
         cmd.env("CARGO_ENCODED_RUSTFLAGS", encode_rustflags(&self.rustflags));
         // Make sure the results end up where we expect them.
+        // Cargo provides multiple ways to adjust this and we need to overwrite all of them.
         let build_target_dir = build_dir.path().join("target");
         cmd.env("CARGO_TARGET_DIR", &build_target_dir);
+        cmd.env("CARGO_BUILD_BUILD_DIR", &build_target_dir);
         // To avoid metadata conflicts, we need to inject some custom data into the crate hash.
         // bootstrap does the same at
         // <https://github.com/rust-lang/rust/blob/c8e12cc8bf0de646234524924f39c85d9f3c7c37/src/bootstrap/builder.rs#L1613>.
@@ -509,7 +528,7 @@ panic = 'unwind'
             TempDir::new_in(&self.sysroot_dir).context("failed to create staging dir")?;
         // Copy the output to `$staging/lib`.
         let staging_lib_dir = staging_dir.path().join("lib");
-        fs::create_dir(&staging_lib_dir).context("faiked to create staging/lib dir")?;
+        fs::create_dir(&staging_lib_dir).context("failed to create staging/lib dir")?;
         let out_dir = build_target_dir
             .join(&target_name)
             .join(DEFAULT_SYSROOT_PROFILE)
@@ -533,16 +552,118 @@ panic = 'unwind'
         .context("failed to write hash file")?;
 
         // Atomic copy to final destination via rename.
-        if sysroot_target_dir.exists() {
-            // Remove potentially outdated files.
-            fs::remove_dir_all(&sysroot_target_dir)
-                .context("failed to clean sysroot target dir")?;
+        // The rename can fail if there was a concurrent build.
+        if fs::rename(staging_dir.path(), sysroot_target_dir).is_err() {
+            // Ensure that that build is identical to what we were going to do.
+            if self.sysroot_read_hash() != Some(cur_hash) {
+                bail!("detected a concurrent sysroot build with different settings");
+            }
         }
-        // Create the *parent* directroy so we can move into it.
-        fs::create_dir_all(&sysroot_target_dir.parent().unwrap())
-            .context("failed to create target directory")?;
-        fs::rename(staging_dir.path(), sysroot_target_dir).context("failed installing sysroot")?;
 
         Ok(SysrootStatus::SysrootBuilt)
     }
 }
+
+/// Collect the patches from the sysroot's workspace `Cargo.toml`.
+fn extract_patches(src_dir: &Path) -> Table {
+    // Assume no patch is needed if the workspace Cargo.toml doesn't exist
+    let workspace_manifest = src_dir.join("Cargo.toml");
+    let f = fs::read_to_string(&workspace_manifest).unwrap_or_else(|e| {
+        panic!(
+            "unable to read workspace manifest at `{}`: {}",
+            workspace_manifest.display(),
+            e
+        )
+    });
+    let mut t: Table = toml::from_str(&f).expect("invalid sysroot workspace Cargo.toml");
+    // We only care about the patch table.
+    t.remove("patch")
+        .map(|v| match v {
+            Value::Table(map) => map,
+            _ => panic!("`patch` is not a table"),
+        })
+        .unwrap_or_default()
+}
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+
+    /// Build a manifest for inspecting, returning the manifest and a source directory.
+    fn setup_manifest_test(config: SysrootConfig) -> (Table, TempDir) {
+        let workspace_toml = r#"
+            [workspace]
+            foo = "bar"
+
+            [patch.crates-io]
+            foo = { path = "bar" }
+            rustc-std-workspace-core = { path = "core" }
+            rustc-std-workspace-alloc = { path = "alloc" }
+            rustc-std-workspace-std = { path = "std" }
+        "#;
+
+        let sysroot = tempfile::tempdir().unwrap(); // dummy sysroot dir, shouldn't be written
+        let src_dir = tempfile::tempdir().unwrap();
+        let f = src_dir.path().join("Cargo.toml");
+        fs::write(&f, workspace_toml).unwrap();
+
+        let builder = SysrootBuilder::new(sysroot.path(), "sometarget").sysroot_config(config);
+        let manifest: Table = toml::from_str(&builder.gen_manifest(src_dir.path())).unwrap();
+        (manifest, src_dir)
+    }
+
+    /// Check that a crate's patch is set to the given path, or that it doesn't exist if `None`.
+    #[track_caller]
+    fn check_patch_path(manifest: &Table, krate: &str, path: Option<&Path>) {
+        let patches = &manifest["patch"]["crates-io"];
+        match path {
+            Some(path) => assert_eq!(
+                &patches[krate]["path"].as_str().unwrap(),
+                &path.to_str().unwrap()
+            ),
+            None => assert!(patches.get(krate).is_none()),
+        }
+    }
+
+    #[test]
+    fn check_patches_no_std() {
+        let (manifest, src_dir) = setup_manifest_test(SysrootConfig::NoStd);
+
+        // Ensure that we use the manifest's paths but mapped to the source directory
+        check_patch_path(&manifest, "foo", Some(&src_dir.path().join("bar")));
+        check_patch_path(
+            &manifest,
+            "rustc-std-workspace-core",
+            Some(&src_dir.path().join("core")),
+        );
+
+        // For `NoStd`, we shouldn't get the alloc and std patches.
+        check_patch_path(&manifest, "rustc-std-workspace-alloc", None);
+        check_patch_path(&manifest, "rustc-std-workspace-std", None);
+    }
+
+    #[test]
+    fn check_patches_with_std() {
+        let (manifest, src_dir) = setup_manifest_test(SysrootConfig::WithStd {
+            std_features: Vec::new(),
+        });
+
+        // For `WithStd`, we should get all patches.
+        check_patch_path(&manifest, "foo", Some(&src_dir.path().join("bar")));
+        check_patch_path(
+            &manifest,
+            "rustc-std-workspace-core",
+            Some(&src_dir.path().join("core")),
+        );
+        check_patch_path(
+            &manifest,
+            "rustc-std-workspace-alloc",
+            Some(&src_dir.path().join("alloc")),
+        );
+        check_patch_path(
+            &manifest,
+            "rustc-std-workspace-std",
+            Some(&src_dir.path().join("std")),
+        );
+    }
+}
